/*  ctg2ppm -- convert CTG file to PPM file with coordinates in comments  */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "basic.h"
#include "pnmops.h"



/* classifications and colors */

typedef struct {		/* info about a CTG classification */
  int code;
  const char *label;
  double red, green, blue;
} CTGInfo;

CTGInfo *ctgInfo(int key);

CTGInfo ctgKey[] = {
  /* 1 Urban or Built-Up Land */
  { 11, "Residential",					0.25, 0.25, 0.25 },
  { 12, "Commercial Services",				0.35, 0.35, 0.35 },
  { 13, "Industrial",					0.45, 0.45, 0.45 },
  { 14, "Transportation, Communications",		1.00, 0.00, 0.00 },
  { 15, "Industrial and Commercial",			0.55, 0.55, 0.55 },
  { 16, "Mixed Urban or Built-Up Land",			0.54, 0.54, 0.54 },
  { 17, "Other Urban or Built-Up Land",			0.53, 0.53, 0.53 },
  /* 2 Agricultural Land */
  { 21, "Cropland and Pasture",				0.50, 0.50, 0.0 },
  { 22, "Orchards, Groves, Vineyards, Nurseries",	0.50, 0.50, 0.1 },
  { 23, "Confined Feeding Operations",			0.50, 0.50, 0.2 },
  { 24, "Other Agricultural Land",			0.50, 0.50, 0.3 },
  /* 3 Rangeland */
  { 31, "Herbaceous Rangeland",				0.81, 0.68, 0.40 },
  { 32, "Shrub and Brush Rangeland",			0.87, 0.81, 0.50 },
  { 33, "Mixed Rangeland",				0.84, 0.75, 0.45 },
  /* 4 Forest Land */
  { 41, "Deciduous Forest Land",			0.00, 0.45, 0.18 },
  { 42, "Evergreen Forest Land",			0.16, 0.45, 0.35 },
  { 43, "Mixed Forest Land",				0.08, 0.40, 0.20 },
  /* 5 Water */
  { 51, "Streams and Canals",				0.01, 0.45, 0.77 },
  { 52, "Lakes",					0.02, 0.45, 0.77 },
  { 53, "Reservoirs",					0.03, 0.45, 0.77 },
  { 54, "Bays and Estuaries",				0.04, 0.45, 0.77 },
  /* 6 Wetland */
  { 61, "Forested Wetlands",				0.23, 0.64, 0.49 },
  { 62, "Nonforested Wetlands",				0.23, 0.60, 0.55 },
  /* 7 Barren Land */
  { 71, "Dry Salt Flats",				0.9, 0.9, 0.9 },
  { 72, "Beaches",					0.9, 0.9, 0.8 },
  { 73, "Sandy Areas Other than Beaches",		0.9, 0.9, 0.7 },
  { 74, "Bare Exposed Rock",				0.9, 0.8, 0.7 },
  { 75, "Strip Mines, Quarries, and Gravel Pits",	0.8, 0.8, 0.7 },
  { 76, "Transitional Areas",				0.8, 0.7, 0.7 },
  { 77, "Mixed Barren Land",				0.9, 0.9, 0.6 },
  /* 8 Tundra */
  { 81, "Shrub and Brush Tundra",			0.7, 0.7, 0.71 },
  { 82, "Herbaceous Tundra",				0.7, 0.7, 0.72 },
  { 83, "Bare Ground",					0.7, 0.7, 0.73 },
  { 84, "Wet Tundra",					0.7, 0.7, 0.74 },
  { 85, "Mixed Tundra",					0.7, 0.7, 0.75 },
  /* 91 Perennial Snowfields */
  { 92, "Glaciers",					1.0, 1.0, 1.0 },
  /* sentinel & default value -- code 0, no label string, dark red */
  {  0, 0, 0.4, 0.0, 0.0 }
};



/*  CTG data structure  */

typedef struct {
  int number_of_rows;
  int number_of_cells;
  int number_of_columns;
  int cell_size;
  int num_overlays;
  int map_type;
  int zone_number;
  int projection_code;
  int scale;
  int source_date;
  int min_col_index, min_row_index;
  int max_col_index, max_row_index;
  int sw_col_index, sw_row_index;
  int nw_col_index, nw_row_index;
  int nc_col_index, nc_row_index;
  int ne_col_index, ne_row_index;
  int se_col_index, se_row_index;
  int sc_col_index, sc_row_index;
  int sw_lat, sw_lon;
  int nw_lat, nw_lon;
  int nc_lat, nc_lon;
  int ne_lat, ne_lon;
  int se_lat, se_lon;
  int sc_lat, sc_lon;
  int nw_utm_easting, nw_utm_northing;
  int creation_date;
  char title[65];
  char *data;
} Ctg;

void initColors(void);
Ctg *readCTG(char *filename);
Ctg *tryReadCTG(FILETYPE in);
int getCTGInt(char *file_record, int start_pos, int end_pos);
void set_ctg_element(Ctg *ctg, int row, int column, int element);
int getCTGLatLon(char *file_record, int start_pos, int end_pos);

void usage(char *progname)
{
  fprintf(stderr,
	  "usage: %s [CTGfile [PPMfile]]\n", progname);
  exit(1);
}

/*  main  */

int main(int argc, char *argv[])
{
  int i,j,n;
  int opti;
  char *inName,*outName;
  FILE *out;
  Ctg *ctg;
  CTGInfo *c;
  UV swUV, nwUV, neUV, seUV;
  UV swXY, nwXY, neXY, seXY;
	
  opti = 1;
  if (argc > opti) {
    inName = argv[opti];
    opti++;
  } else {
    inName = NULL;
  }
  if (argc > opti) {
    outName = argv[opti];
    opti++;
  } else {
    outName = NULL;
  }
  if (argc > opti) usage(argv[0]);

  if (outName) {
    out = fopen(outName,"wb");
    if (out==NULL) {
      fprintf(stderr, "Can't open %s for write.\n",outName);
      exit(1);
    }
  } else {
    out = stdout;
  }
	
  initColors();		/* load mapping table */
	
  ctg = readCTG(inName);

  if (ctg == NULL)	/* Error reading CTG. */
    exit(1);

  swXY.u = ctg->sw_col_index - ctg->min_col_index;
  swXY.v = ctg->max_row_index - ctg->sw_row_index;
  nwXY.u = ctg->nw_col_index - ctg->min_col_index;
  nwXY.v = ctg->max_row_index - ctg->nw_row_index;
  neXY.u = ctg->ne_col_index - ctg->min_col_index;
  neXY.v = ctg->max_row_index - ctg->ne_row_index;
  seXY.u = ctg->se_col_index - ctg->min_col_index;
  seXY.v = ctg->max_row_index - ctg->se_row_index;
	
  swUV.u = ctg->sw_lon;
  swUV.v = ctg->sw_lat;
  nwUV.u = ctg->nw_lon;
  nwUV.v = ctg->nw_lat;
  neUV.u = ctg->ne_lon;
  neUV.v = ctg->ne_lat;
  seUV.u = ctg->se_lon;
  seUV.v = ctg->se_lat;
	
  /* Print out the PPM file */

  fprintf(out,"P6\n");
  fprintf(out,TOPOFLAG);
  fprintf(out,LEGENDFLAG);
  /* number of entries minus sentinel */
  n = sizeof(ctgKey)/sizeof(ctgKey[0]) - 1;
  fprintf(out,"# %d\n",n);
  for(i=0;i<n;++i) {
    fprintf(out,"#%s#\t%f %f %f\n",
	    ctgKey[i].label,ctgKey[i].red,ctgKey[i].green,ctgKey[i].blue);
  }
  fprintf(out,REFPTHEADER);
  fprintf(out,REFPTFORMATOUT,swUV.u/3600., swUV.v/3600.,swXY.u, swXY.v);
  fprintf(out,REFPTFORMATOUT,nwUV.u/3600., nwUV.v/3600.,nwXY.u, nwXY.v);
  fprintf(out,REFPTFORMATOUT,neUV.u/3600., neUV.v/3600.,neXY.u, neXY.v);
  fprintf(out,REFPTFORMATOUT,seUV.u/3600., seUV.v/3600.,seXY.u, seXY.v);
  fprintf(out,"%d %d\n", ctg->number_of_columns, ctg->number_of_rows);
  fprintf(out,"255\n");
	
  for (i = 0; i < ctg->number_of_rows; i++) {
    for (j = 0; j <ctg->number_of_columns; j++) {
      c = ctgInfo((int)(*(ctg->data + i * ctg->number_of_columns + j)));
      fprintf(out,"%c%c%c", (unsigned char)(c->red * 255.0),
	      (unsigned char)(c->green * 255.0),
	      (unsigned char)(c->blue * 255.0));
    }
  }
  return 0;
}

/*  initColors()  -- initialize color mapping table  */

#define CTGMAX 100			/* maximum key value plus 1 */
CTGInfo *ctgMap[CTGMAX];		/* map from key to label and color */

void initColors(void)
{
  int i;
  CTGInfo *p;
	
  /* preload with pointers to sentinel/filler entry */
  i = sizeof(ctgKey) / sizeof(ctgKey[0]);	/* number of entries */
  p = ctgKey + i - 1;				/* pointer to last entry */
  for (i = 0; i < CTGMAX; i++)
    ctgMap[i] = p;
	
  /* load recognized values from table above */
  for (p = ctgKey; p->code != 0; p++)
    ctgMap[p->code] = p;
}



/*  ctgInfo(key) -- return pointer to Info struct for key  */

CTGInfo *ctgInfo(int key)
{
  if (key >= 0 && key < CTGMAX)	/* if in range */
    return ctgMap[key];		/* return corresponding entry */
  else
    return ctgMap[0];		/* else return first entry */
}



/* CTG processing code */


/*--------------------------getCTGInt()--------------------------*/

int getCTGInt(char *file_record, int start_pos, int end_pos)
{
  char tmp_str[100];
	
  memcpy(tmp_str, file_record + start_pos - 1, end_pos - start_pos + 1);
  tmp_str[end_pos-start_pos + 1] = 0;
	
  return atoi(tmp_str);
}

/*--------------------------getCTGLatLon()--------------------------*/

int getCTGLatLon(char *file_record, int start_pos, int end_pos)
{
  char tmp_str[100];
  int degrees, minutes, seconds;
	
  memcpy(tmp_str, file_record + start_pos - 1, 10);
  tmp_str[10] = 0;
  seconds = atoi(tmp_str + 8);
  tmp_str[8] = 0;
  minutes = atoi(tmp_str + 6);
  tmp_str[6] = 0;
  degrees = atoi(tmp_str);
	
  return seconds + (minutes + degrees * 60) * 60;
}

/*--------------------------set_ctg_element()--------------------------*/

void set_ctg_element(Ctg *ctg, int row, int column, int element)
{
  *(ctg->data + row * ctg->number_of_columns + column) = element;
}

/*-----------------------tryReadCTG()--------------------------*/
Ctg *tryReadCTG(FILETYPE in)
{
  char file_record[81];
  int i, j;
  int records_read, percent_recs_read;
  int x, y, column, row;
  Ctg *ctg;
	
  ctg = (Ctg *)malloc(sizeof(Ctg));
  if (ctg == NULL) {
    fprintf(stderr,"Can't allocate space for Ctg structure.\n");
    return NULL;
  }

  /* RECORD 1 */
  if (!FREAD(file_record, 80, 1, in)) {
    fprintf(stderr, "Error reading CTG file.\n");
    free(ctg);
    return NULL;
  }
  ctg->number_of_rows = getCTGInt(file_record, 1, 10);
  ctg->number_of_cells = getCTGInt(file_record, 11, 20) / 2;
  ctg->number_of_columns = getCTGInt(file_record, 21, 30);
	
  ctg->cell_size = getCTGInt(file_record, 36, 40);
  ctg->num_overlays = getCTGInt(file_record, 41, 45);
  ctg->map_type = getCTGInt(file_record, 46, 50);
  ctg->zone_number = getCTGInt(file_record, 51, 55);
  ctg->projection_code = getCTGInt(file_record, 56, 60);
  ctg->scale = getCTGInt(file_record, 61, 70);
  ctg->source_date = getCTGInt(file_record, 61, 70);
  if (ctg->number_of_rows == 0 ||
      ctg->number_of_cells == 0 ||
      ctg->number_of_columns == 0 ||
      ctg->cell_size == 0) {
    fprintf(stderr, "Error reading CTG file.\n");
    free(ctg);
    return NULL;
  }
    
	
  /* RECORD 2 */
  if (!FREAD(file_record, 80, 1, in)) {
    fprintf(stderr, "Error reading CTG file.\n");
    free(ctg);
    return NULL;
  }
  ctg->min_col_index = getCTGInt(file_record, 1, 5);
  ctg->min_row_index = getCTGInt(file_record, 6, 10);
  ctg->max_col_index = getCTGInt(file_record, 11, 15);
  ctg->max_row_index = getCTGInt(file_record, 16, 20);
  ctg->sw_col_index = getCTGInt(file_record, 21, 25);
  ctg->sw_row_index = getCTGInt(file_record, 26, 30);
  ctg->nw_col_index = getCTGInt(file_record, 31, 35);
  ctg->nw_row_index = getCTGInt(file_record, 36, 40);
  ctg->nc_col_index = getCTGInt(file_record, 41, 45);
  ctg->nc_row_index = getCTGInt(file_record, 46, 50);
  ctg->ne_col_index = getCTGInt(file_record, 51, 55);
  ctg->ne_row_index = getCTGInt(file_record, 56, 60);
  ctg->se_col_index = getCTGInt(file_record, 61, 65);
  ctg->se_row_index = getCTGInt(file_record, 66, 70);
  ctg->sc_col_index = getCTGInt(file_record, 71, 75);
  ctg->sc_row_index = getCTGInt(file_record, 76, 80);
	
  /* RECORD 3 */
  if (!FREAD(file_record, 80, 1, in)) {
    fprintf(stderr, "Error reading CTG file.\n");
    free(ctg);
    return NULL;
  }
  ctg->sw_lat = getCTGLatLon(file_record, 1, 10);
  ctg->sw_lon = getCTGLatLon(file_record, 11, 20);
  ctg->nw_lat = getCTGLatLon(file_record, 21, 30);
  ctg->nw_lon = getCTGLatLon(file_record, 31, 40);
  ctg->nc_lat = getCTGLatLon(file_record, 41, 50);
  ctg->nc_lon = getCTGLatLon(file_record, 51, 60);
  ctg->ne_lat = getCTGLatLon(file_record, 61, 70);
  ctg->ne_lon = getCTGLatLon(file_record, 71, 80);
	
  /* RECORD 4 */
  if (!FREAD(file_record, 80, 1, in)) {
    fprintf(stderr, "Error reading CTG file.\n");
    free(ctg);
    return NULL;
  }
  ctg->se_lat = getCTGLatLon(file_record, 1, 10);
  ctg->se_lon = getCTGLatLon(file_record, 11, 20);
  ctg->sc_lat = getCTGLatLon(file_record, 21, 30);
  ctg->sc_lon = getCTGLatLon(file_record, 31, 40);
  /* Fix unsigned longitudes (grrr...) */
  if (ctg->zone_number < 31) {
    ctg->sw_lon = -ABS(ctg->sw_lon);
    ctg->nw_lon = -ABS(ctg->nw_lon);
    ctg->nc_lon = -ABS(ctg->nc_lon);
    ctg->ne_lon = -ABS(ctg->ne_lon);
    ctg->se_lon = -ABS(ctg->se_lon);
    ctg->sc_lon = -ABS(ctg->sc_lon);
  }
  ctg->nw_utm_easting = getCTGInt(file_record, 41, 50);
  ctg->nw_utm_northing = getCTGInt(file_record, 51, 60);
  ctg->creation_date = getCTGInt(file_record, 61, 70);
	
  /* RECORD 5 */
  if (!FREAD(file_record, 80, 1, in)) {
    fprintf(stderr, "Error reading CTG file.\n");
    free(ctg);
    return NULL;
  }
	
  memcpy(ctg->title, file_record, 64);
  ctg->title[64] = 0;
	
  ctg->data = malloc(ctg->number_of_rows * ctg->number_of_columns);
  if (!ctg->data){
    fprintf(stderr,"Can't allocate space for Ctg structure.\n");
    return NULL;
  }
	
  for (i = 0; i < ctg->number_of_rows; i++) {
    for (j = 0; j < ctg->number_of_columns; j++) {
      set_ctg_element(ctg, i, j, 0);
    }
  }
	
  records_read = 0;
  percent_recs_read = -1;
  while (FREAD(file_record, 80, 1, in)) {
    x = getCTGInt(file_record, 4, 11);
    y = getCTGInt(file_record, 12, 19);
    column = (x - ctg->nw_utm_easting)/ctg->cell_size + 1 - ctg->min_col_index;
    row = (ctg->nw_utm_northing - y)/ctg->cell_size + 1 - ctg->min_row_index;
    if (column >= ctg->number_of_columns || column < 0) {
      fprintf(stderr, "Easting out of range (%d meters)\n", x);
      free(ctg);
      return NULL;
    }
    if (row >= ctg->number_of_rows || row < 0) {
      fprintf(stderr, "Northing out of range (%d meters)\n", y);
      free(ctg);
      return NULL;
    }
    set_ctg_element(ctg, row, column, getCTGInt(file_record, 21, 30));
    records_read++;
    if (percent_recs_read != records_read * 100 / ctg->number_of_cells) {
      percent_recs_read = (records_read * 100 / ctg->number_of_cells);
      fprintf(stderr, "\rReading CTG file... %d%%", percent_recs_read);
      fflush(stdout);
    }
  }
	
  fprintf(stderr, "\n");
  return(ctg);
}

/*--------------------------readCTG()--------------------------*/
Ctg *readCTG(char *filename)
{
  Ctg *ctg;
  FILETYPE f;

  if( (f = open4read(filename)) == NULL ) return NULL;
  ctg = tryReadCTG(f);
  FCLOSE(f);
  return ctg;
}
