/*  dem.c -- functions for handling USGS Digital Elevation Model files  */

#include "dem.h"

static Dem *tryReadDEM(FILETYPE f);
static Em *dem2em(Dem *dem);
static int getDEMInt(FILETYPE f);
static int getDEMInt2(FILETYPE f);
static int getDEMInt6(FILETYPE f);
static double getDEMDouble12(FILETYPE f);
static double getDEMDouble24(FILETYPE f);

void transEMxy2lola(Em *em, UV xy, UV *lola)
{
  UV utm;

  switch( em->ref_system ) {
  case UTM_SYSTEM:
    utm.u = xy.u * em->groundDelta.u + em->groundO.u;
    utm.v = xy.v * em->groundDelta.v + em->groundO.v;
    utm2lola(em->zone, utm, lola);
    break;
  case GEO_SYSTEM:
    lola->u = xy.u * em->groundDelta.u + em->groundO.u;
    lola->v = xy.v * em->groundDelta.v + em->groundO.v;
    break;
  default:
    fprintf(stderr, "Error: Unknown reference system.\n");
    exit(-1);
  }
}

void transEMlola2xy(Em *em, UV lola, UV *xy)
{
  UV utm;

  switch( em->ref_system ) {
  case UTM_SYSTEM:
    lola2utm(lola, &utm);
    xy->u = (utm.u - em->groundO.u) / em->groundDelta.u;
    xy->v = (utm.v - em->groundO.v) / em->groundDelta.v;
    break;
  case GEO_SYSTEM:
    xy->u = (lola.u - em->groundO.u) / em->groundDelta.u;
    xy->v = (lola.v - em->groundO.v) / em->groundDelta.v;
    break;
  default:
    fprintf(stderr, "Error: Unknown reference system.\n");
    exit(-1);
  }
}

Em *readEM(char *filename)
{
  Dem *dem;
  Em *em;

  dem = readDEM(filename);
  em = dem2em(dem);
  free(dem);
  return(em);
}


Dem *readDEM(char *filename)
{
  Dem *dem;
  FILETYPE f;

  if( (f = open4read(filename)) == NULL ) return NULL;
  dem = tryReadDEM(f);
  FCLOSE(f);
  return dem;
}


static Em *dem2em(Dem *dem)
{
  Em *em;
  int i, j, length;
  unsigned short *p;
  UV min;
	
  if (dem == NULL) return NULL;

  em = (Em *)malloc(sizeof(Em));
  if (!em) nospace();
	
  em->width = dem->cols;
  em->height = dem->rows;
  length = em->width * em->height;
  em->data = (unsigned short *)malloc(sizeof(short) * length);
  /*modified by Xiangyu Zhang*/
  em->normal=(float *) malloc(sizeof(float)*length*3);
  
  if (!em->data||!em->normal) nospace();
  /**/
  
  em->dz2m = dem->dz2m;
  em->dzoffset = dem->dzoffset;
  em->dzmax = dem->dzmax;

  for (i = 0; i < 4; i++) {
    if (i == 0) {
      min.u = dem->corner[i].u;
      min.v = dem->corner[i].v;
    } else {
      min.u = MIN(min.u, dem->corner[i].u);
      min.v = MIN(min.v, dem->corner[i].v);
    }
  }
  /* For accurate transformation from xy to lola */
  em->zone = dem->zone;
  em->datum = dem->datum;
  em->ref_system = dem->reference_system;

  /* the units will be either arcseconds or UTM, depending on ref_system */
  if( em->ref_system == UTM_SYSTEM ) {
    /* grid starts at UTM coord that is a multiple of the resolution. */
    em->groundO.u = ceil(min.u / dem->resolution[0]) * dem->resolution[0];
    em->groundO.v = ceil(min.v / dem->resolution[1]) * dem->resolution[1];
    /* determine datum */
  } else {
    em->groundO.u = min.u;
    em->groundO.v = min.v;
  }
  em->groundDelta.u = dem->resolution[0];
  em->groundDelta.v = dem->resolution[1];


  em->swXY.u = floor((dem->corner[0].u - min.u) / dem->resolution[0]);
  if (em->swXY.u < 0.0)
    em->swXY.u = 0.0;
  em->swXY.v = floor((dem->corner[0].v - min.v) / dem->resolution[1]);
  if (em->swXY.v < 0.0)
    em->swXY.v = 0.0;
  em->nwXY.u = floor((dem->corner[1].u - min.u) / dem->resolution[0]);
  if (em->nwXY.u < 0.0)
    em->nwXY.u = 0.0;
  em->nwXY.v = floor((dem->corner[1].v - min.v) / dem->resolution[1]);
  if (em->nwXY.v > dem->rows - 1)
    em->nwXY.v = dem->rows - 1;
  em->neXY.u = floor((dem->corner[2].u - min.u) / dem->resolution[0]);
  if (em->neXY.u > dem->cols - 1)
    em->neXY.u = dem->cols - 1;
  em->neXY.v = floor((dem->corner[2].v - min.v) / dem->resolution[1]);
  if (em->neXY.v > dem->rows - 1)
    em->neXY.v = dem->rows - 1;
  em->seXY.u = floor((dem->corner[3].u - min.u) / dem->resolution[0]);
  if (em->seXY.u > dem->cols - 1)
    em->seXY.u = dem->cols - 1;
  em->seXY.v = floor((dem->corner[3].v - min.v) / dem->resolution[1]);
  if (em->seXY.v < 0.0)
    em->seXY.v = 0.0;
	
  if (dem->reference_system == 1 && dem->ground_unit == 2) {
		
    /* coordinate system in UTM with meter ground units */
		
    utm2lola(dem->zone, dem->corner[0], &(em->swUV));
    utm2lola(dem->zone, dem->corner[1], &(em->nwUV));
    utm2lola(dem->zone, dem->corner[2], &(em->neUV));
    utm2lola(dem->zone, dem->corner[3], &(em->seUV));
  } else {
		
    /* Hope the coordinate system is in arcseconds */
		
    em->swUV.u = dem->corner[0].u;
    em->swUV.v = dem->corner[0].v;
    em->nwUV.u = dem->corner[1].u;
    em->nwUV.v = dem->corner[1].v;
    em->neUV.u = dem->corner[2].u;
    em->neUV.v = dem->corner[2].v;
    em->seUV.u = dem->corner[3].u;
    em->seUV.v = dem->corner[3].v;
  }
	
  p = em->data;
  for(j = 0; j < dem->rows; ++j) {
    for(i = 0; i < dem->cols; ++i) {
      *p++ = dem->data[i][j];
    }
  }
  return(em);
}




static Dem *tryReadDEM(FILETYPE f)
{
  Dem *dem;
  int i,j,k;
  int col;
  int row_id, col_id;
  int colsInProfiles, numProfiles;
  UV min, max;
  int firstIndex;
  int dzoffset;
  int truncated;
  int tmp;
  unsigned short dzmax;

  dem = (Dem *)malloc(sizeof(Dem));
  if (dem == NULL) {
    fprintf(stderr,"Can't allocate space for Dem structure.\n");
    return(NULL);
  }
	
  /* A records */
	
  /* A 1 */
  if (!FGETS(dem->name, 145, f)) {
    fprintf(stderr, "Empty input file.\n");
    free(dem);
    return NULL;
  }
	
  /* A 2 */
  dem->level = getDEMInt6(f);
	
  /* A 3 */
  dem->elevation_pattern = getDEMInt6(f);
	
  /* A 4 */
  dem->reference_system = getDEMInt6(f);

  /* A 5 */
  dem->zone = getDEMInt6(f);
	
  /* A 6 */
  for (i = 0; i < 15; i++) {
    dem->map_projection_parameter[i] = getDEMDouble24(f);
  }
	
  /* A 7 */
  dem->ground_unit = getDEMInt6(f);
	
  /* A 8 */
  dem->elevation_unit = getDEMInt6(f);
  if (dem->elevation_unit == 1) {
    dem->dz2m = METERS_PER_FEET;
  } else {
    dem->dz2m = 1.0;
  }
	
  /* A 9 */
  dem->n_polygon = getDEMInt6(f);
	
  /* A 10 */
  for (i = 0; i < 4; i++) {
    dem->corner[i].u = getDEMDouble24(f);
    dem->corner[i].v = getDEMDouble24(f);
    if (i == 0) {
      min.u = dem->corner[i].u;
      max.u = dem->corner[i].u;
      min.v = dem->corner[i].v;
      max.v = dem->corner[i].v;
    } else {
      min.u = MIN(min.u, dem->corner[i].u);
      max.u = MAX(max.u, dem->corner[i].u);
      min.v = MIN(min.v, dem->corner[i].v);
      max.v = MAX(max.v, dem->corner[i].v);
    }
  }
	
  /* A 11 */
  dem->minElevation = getDEMDouble24(f);
  dem->maxElevation = getDEMDouble24(f);

  /* A 12 */
  dem->axis_angle = getDEMDouble24(f);		/* assumed to be 0.0 */
	
  /* A 13 */
  dem->accuracy = getDEMInt6(f);
	
  /* A 14 */
  for (i = 0; i < 3; i++) {
    dem->resolution[i] = getDEMDouble12(f);
  }
  if (dem->resolution[2] == 0) {
    fprintf(stderr,"Error reading DEM file.\n");
    free(dem);
    return NULL;
  }
  dem->dz2m *= dem->resolution[2];	/* scale by accuracy */
  dzoffset = (int)(dem->minElevation/dem->resolution[2])-1;
  if (dzoffset < -32767) dzoffset=0-1;
	
  /* A 15 */
  colsInProfiles = getDEMInt6(f);
  if (colsInProfiles != 1) {
    fprintf(stderr, "Cannot handle more than one column profiles.\n");
    free(dem);
    return NULL;
  }

  numProfiles = getDEMInt6(f);
  if (numProfiles > MAXCOLS) {
    fprintf(stderr, "Too many profiles (%d) in data set.\n",numProfiles);
    free(dem);
    return NULL;
  }
	
  /* A remaining */
  /*
    All USGS DEM's lacking the new type A record, elements 16-29, are NAD 27.
    All NIMA DEM's and DTED's lacking datum descriptors are WGS 72.
    All DEM's having the new type A record elements have datums as
    indicated in type A record, element 27.

    1=North American Datum 1927 (NAD 27)
    2=World Geodetic System 1972 (WGS 72)
    3=WGS 84
    4=NAD 83
    5=Old Hawaii Datum
    6=Puerto Rico Datum

    Since I don't know how to determine a NIMA DEM or DTED, I assume NAD 27
    (for UTM) and WGS 72 (for GEO) unless told otherwise.
  */
  if( dem->reference_system == UTM_SYSTEM ) {
    dem->datum = NAD27;
  } else {
    dem->datum = WGS72;
  }
  for (i = 865; i <= 1024; i++) {
    if( i == 891 ) {
      j = getDEMInt2(f);
      if( j > 0 ) {
	dem->datum = j;
      }
      i += 1;
    } else if (GETC(f) == '\n')
      break;
  }

  /* B records */
		
  dem->cols = numProfiles;
  dem->rows = (int)(1 + (max.v - min.v)/dem->resolution[1]);
		
  if (dem->rows > MAXROWS) {
    fprintf(stderr, "Too many rows (%d) in DEM.\n",dem->rows);
    free(dem);
    return NULL;
  }

  truncated = 0;
  dzmax = 0;
  for (col = 0; col < numProfiles; ++col) {
    for (i = 0; i < dem->rows; ++i) {
      dem->data[col][i] = 0;
    }
			
    /* B 1 */
    row_id = getDEMInt(f);
    col_id = getDEMInt(f);
    if (col_id != col + 1 || row_id != 1) {
      fprintf(stderr, "Profile %d claims to be %d.\n", col + 1, col_id);
      free(dem);
      return NULL;
    }
			
    /* B 2 */
    dem->rowsInProfile[col] = getDEMInt6(f);
    dem->colsInProfile[col] = getDEMInt6(f);
			
    /* B 3 */
    dem->firstUVInProfile[col].u = getDEMDouble24(f);
    dem->firstUVInProfile[col].v = getDEMDouble24(f);
    firstIndex = (int)((dem->firstUVInProfile[col].v - min.v)/dem->resolution[1]);
			
    /* B 4 */
    dem->firstElevationInProfile[col] = getDEMDouble24(f);
			
    /* B 5 */
    dem->minElevationInProfile[col] = getDEMDouble24(f);
    dem->maxElevationInProfile[col] = getDEMDouble24(f);
			
    /* B 6 */
#ifdef DEBUG
    fprintf(stderr, "Profile %d starts with row (%lf - %lf)/%lf = %d.\n",
	    col + 1, dem->firstUVInProfile[col].v, min.v, dem->resolution[1],
	    firstIndex);
#endif
			
    for (i = firstIndex; i < firstIndex + dem->rowsInProfile[col]; i++) {
      if (i >= dem->rows) {
	fprintf(stderr, "Too many rows in DEM profile %d.\n", col + 1);
	free(dem);
	return NULL;
      }

      /* WARNING some DEM files have record delimiters (like '\n')
	 that may break a single elevation into two.  We do not handle
	 these types of DEM files.
	 */
      tmp = getDEMInt(f);

      if (tmp != -32767) {		/* -32767 ==> no data */
	if (tmp < dzoffset+1) {		/* the min elevation was wrong */
#ifdef DEBUG
	  fprintf(stderr,"Rescaling for new minimum elevation.\n");
#endif
	  for (j = 0; j <= col; ++j) {
	    for (k = 0; k < dem->rows; ++k) {
	      if (dem->data[j][k] != 0) {
		if (dem->data[j][k] + dzoffset-tmp+1 > USHRT_MAX) {
		  truncated = 1;
		  dem->data[j][k] = USHRT_MAX;
		} else {
		  dem->data[j][k] += (unsigned short)(dzoffset-tmp+1);
		}
		dzmax = MAX(dzmax,dem->data[j][k]);
	      }
	    }
	  }
	  dzoffset = tmp-1;
	}

	if (tmp-dzoffset > USHRT_MAX) {
	  truncated = 1;
	  dem->data[col][i] = USHRT_MAX;
	} else {
	  dem->data[col][i] = (unsigned short)(tmp-dzoffset);
	}
	dzmax = MAX(dzmax,dem->data[col][i]);
      }
    }
			
  }
		
  if (truncated) {
    fprintf(stderr,"**Warning** Some elevations have been truncated.\n");
  }
  dem->dzoffset = dzoffset;
  dem->dzmax = dzmax;

  /* C record */
		
  if (dem->accuracy != 0) {
			
    /* C 1 */
    dem->C2 = getDEMInt6(f);
			
    /* C 2 */
    dem->C2x = getDEMInt6(f);
    dem->C2y = getDEMInt6(f);
    dem->C2z = getDEMInt6(f);
			
    /* C 3 */
    dem->C2sample_size = getDEMInt6(f);
			
    /* C 4 */
    dem->C5 = getDEMInt6(f);
			
    /* C 5 */
    dem->C5x = getDEMInt6(f);
    dem->C5y = getDEMInt6(f);
    dem->C5z = getDEMInt6(f);
			
    /* C 6 */
    dem->C5sample_size = getDEMInt6(f);
  }
		
  return(dem);
}


static void getDEMD2E(int n, char *buf, FILETYPE f)
{
  int i;
	
  for (i = 0; i < n; ++i) {
    buf[i] = GETC(f);
    if (buf[i] == 'D') {
      buf[i] = 'E';
    }
  }
  buf[n] = '\0';
}



/* ? characters */
static int getDEMInt(FILETYPE f)
{
  int i = 0, c;
  int sgn = 1;
	
  while( (c = GETC(f)) != EOF && isspace(c) )
    ;
  if( c == '-' ) {
    sgn = -1;
    c = GETC(f);
  } else if( c == '+' ) {
    c = GETC(f);
  }
  while( isdigit(c) ) {
    i = 10 * i + (c - '0');
    c = GETC(f);
  }
  UNGETC(c,f);
  return sgn * i;
}


static void getDEMNoCR(int n, char *buf, FILETYPE f)
{
  int i;
	
  for (i = 0; i < n; ++i) {
    buf[i] = GETC(f);
    if (buf[i] == '\n') {
      --i;
    }
  }
  buf[n] = '\0';
	
}


/* 2 characters */
static int getDEMInt2(FILETYPE f)
{
  char buf[3];
	
  getDEMNoCR(2, buf, f);
  return atoi(buf);
}


/* 6 characters */
static int getDEMInt6(FILETYPE f)
{
  char buf[7];
	
  getDEMNoCR(6, buf, f);
  return atoi(buf);
}


/* 12 characters */
static double getDEMDouble12(FILETYPE f)
{
  char buf[13];
	
  getDEMD2E(12, buf, f);
  return atof(buf);
}



/* 24 characters */
static double getDEMDouble24(FILETYPE f)
{
  char buf[25];
	
  getDEMD2E(24, buf, f);
  return atof(buf);
}
