/*  emit.c -- draw a surface  */

#ifdef WIN32
#include <windows.h>
#endif
#include "topo.h"
#include <GL/gl.h>

static void emitDFS(TriIndex t, Coord2d p1, Coord2d p2, Coord2d p3);
static void emitColor(int x, int y, int z);
static void emitTri(TriIndex t, Coord2d p1, Coord2d p2, Coord2d p3);



static int surfType;		/* surface type: WIRE2D, WIRE3D, or FILL3D */
static double zscale;
static Hier *hp;

/*  emitSurface(hp, type) -- draw surface from triangle hierarchy  */

void emitSurface(Hier *h, int type)
{
  hp = h;
  surfType = type;
  zscale = vexagger * m2grid * gem->dz2m;
  numDisplayTri = 0;
  if (surfType == WIRE2D) {
    glPolygonMode(GL_FRONT_AND_BACK,GL_LINE);
    glColor3f(0.0,0.0,0.0);
  } else if (surfType == FOV2D) {
    glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);
  } else if (surfType == WIRE3D) {
    glPolygonMode(GL_FRONT_AND_BACK,GL_LINE);
    glColor3f(0.0,0.0,0.0);
  } else {	/* FILL3D */
    glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);
  }
  glBegin(GL_TRIANGLES);
  emitDFS(LCHILD(0), sw, ne, nw);	/* NW corner of square */
  emitDFS(RCHILD(0), ne, sw, se);	/* SE corner of square */
  glEnd();
}


/*  emitDFS(t, p1, p2, p3) -- emit triangle, possibly subdivided  */

static void emitDFS(TriIndex t, Coord2d p1, Coord2d p2, Coord2d p3)
{
  Coord2d mid;

  if (ONSURFACE(hp,t)) {
    emitTri(t, p1, p2, p3);
  } else {
		
    assert(!ISLEAF(hp, t));
		
    MIDPOINT(p1, p2, mid);
    emitDFS(LCHILD(t), p3, p1, mid);
    emitDFS(RCHILD(t), p2, p3, mid);
  }
}





#define GET_NORMAL {				\
  nx = (y2-y1)*(sz3-sz1) - (sz2-sz1)*(y3-y1);	\
  ny = (sz2-sz1)*(x3-x1) - (x2-x1)*(sz3-sz1);	\
  nz = (GLfloat)((x2-x1)*(y3-y1) - (y2-y1)*(x3-x1));	\
}

/*  emitTri(t, p1, p2, p3) -- emit one triangle  */
static void emitTri(TriIndex t, Coord2d p1, Coord2d p2, Coord2d p3)
{
  int x1,x2,x3,y1,y2,y3,z1,z2,z3;
  GLfloat sz1, sz2, sz3;
  GLfloat nx, ny, nz;

  if (INROD(hp,t) && (!cullFOV || INFOV(hp,t))) {
			
    x1=p1[0];
    y1=p1[1];
    z1=EMDATA(gem, x1, y1);
    sz1 = (GLfloat)zscale*z1;

    x2=p2[0];
    y2=p2[1];
    z2=EMDATA(gem, x2, y2);
    sz2 = (GLfloat)zscale*z2;

    x3=p3[0];
    y3=p3[1];
    z3=EMDATA(gem, x3, y3);
    sz3 = (GLfloat)zscale*z3;

    numDisplayTri++;

    if (surfType == WIRE2D) {
      glVertex2i(x1, y1);
      glVertex2i(x2, y2);
      glVertex2i(x3, y3);
    } else if (surfType == FOV2D) {
      switch(INFOV(hp,t)) {
      case ALLINFOV:
	glColor3f(1.0, 1.0, 0.0);
	break;
      case PARTINFOV:
	glColor3f(1.0, 0.0, 0.0);
	break;
      default:
	glColor3f(0.5, 0.5, 0.5);
      }
      glVertex2i(x1, y1);
      glVertex2i(x2, y2);
      glVertex2i(x3, y3);
    } else if (surfType == WIRE3D) {
      glVertex3f((GLfloat)x1, (GLfloat)y1, sz1);
      glVertex3f((GLfloat)x2, (GLfloat)y2, sz2);
      glVertex3f((GLfloat)x3, (GLfloat)y3, sz3);
    } else {	/* FILL3D */
      if( smoothSurface ) {
	nx=EMNORMAL(gem,x1,y1,0);
	ny=EMNORMAL(gem,x1,y1,1);
	nz=EMNORMAL(gem,x1,y1,2);
	glNormal3f(nx, ny, nz);
	emitColor(x1, y1, z1); glVertex3f((GLfloat)x1, (GLfloat)y1, sz1);
      
	nx=EMNORMAL(gem,x2,y2,0);
	ny=EMNORMAL(gem,x2,y2,1);
	nz=EMNORMAL(gem,x2,y2,2);
	glNormal3f(nx, ny, nz);
	emitColor(x2, y2, z2); glVertex3f((GLfloat)x2, (GLfloat)y2, sz2);
      
	nx=EMNORMAL(gem,x3,y3,0);
	ny=EMNORMAL(gem,x3,y3,1);
	nz=EMNORMAL(gem,x3,y3,2);
	glNormal3f(nx, ny, nz);
	emitColor(x3, y3, z3); glVertex3f((GLfloat)x3, (GLfloat)y3, sz3);
      } else {
	GET_NORMAL;
	glNormal3f(nx, ny, nz);
	emitColor(x1, y1, z1); glVertex3f((GLfloat)x1, (GLfloat)y1, sz1);
	emitColor(x2, y2, z2); glVertex3f((GLfloat)x2, (GLfloat)y2, sz2);
	emitColor(x3, y3, z3); glVertex3f((GLfloat)x3, (GLfloat)y3, sz3);
      }      
    }
  }
}


static void emitColor(int x, int y, int z)
     /* emit color depending on location and altitude  */
{
  if (colorLayer == T_CONTOURS){	/* base the color on the elevation */
    Color *c;

    if (z == 0) {
      glColor3f(0.4f, 0.0f, 0.0f);		/* use dark red for non-data */
    } else if (z <= high_water && z >= low_water) {
      glColor3f(0.0f, 0.0f, 0.4f);		/* use dark blue for sea level */
    } else {
      c = elevMap + (int)(z * dz2contour);
      glColor3f(c->red, c->green, c->blue);
    }
			
  } else if (colorLayer == T_GRAY) {	/* gray with slight tan tint */
			
    glColor3f(0.6f, 0.6f, 0.55f);
			
  } else {				/* use color from PPM files */

    glColor3ub(RED(clayer[colorLayer], x, y),
	      GREEN(clayer[colorLayer], x, y),
	      BLUE(clayer[colorLayer], x, y));
  }			
}
